@@ -21,6 +21,8 @@ use apexl\Io\modules\subscription\entities\subscriptionCreditEntity;
 use apexl\Io\modules\subscription\entities\subscriptionEntity;
 use apexl\Io\modules\subscription\services\subscriptionService;
 use apexl\Io\modules\user\entities\userEntity;
+use apexl\Io\modules\user\services\currentUser;
+use apexl\Io\services\Database;
 use apexl\Io\services\Logger;
 use apexl\Io\services\Mailer;
 use apexl\Vault\Vault;
@@ -116,19 +118,32 @@ class invoiceService {
                     'emailAddress' => $userEntity->email,
                     'taxNumber' => ""
                 ];
+                    if (isset($userEntity->invoice_name_override) && trim($userEntity->invoice_name_override) != "") {
+                        $contactData->name = $userEntity->invoice_name_override;
+                        $lookupContactBy = 'name';
+                    }
                 }
 
                 if (!$brandingThemeId && isset($this->config->app->invoicing->xero->branding_themes->default)) {
                     $brandingThemeId = $this->config->app->invoicing->xero->branding_themes->default;
                 }
                 $brandingTheme = $brandingThemeId ?? $this->config->app->invoicing->xero->default_branding_theme;
-                list($invoiceId, $invoiceNumber, $contactData) = $this->createProviderInvoice($lineItems, $contactData, $lookupContactBy, null, '', $brandingTheme, $sendEmail, $invReference, 'ACCREC', 'GBP', $lineAmountType);
+
+                try {
+                    list($invoiceId, $invoiceNumber, $contactData) = $this->createProviderInvoice($lineItems, $contactData, $lookupContactBy, null, '', $brandingTheme, $sendEmail, $invReference, 'ACCREC', 'GBP', $lineAmountType);
 
                 $this->providerContactData[$invoiceNumber] = $contactData;
 
                 $invoice->provider_id = $invoiceId;
                 $invoice->invoice_number = $invoiceNumber;
                 $invoice->store();
+                } catch (\Exception $e) {
+                    $invoice->provider_error = $e->getMessage();
+                    $invoice->on_hold = 1;
+                    $invoice->store();
+
+                    $this->providerErrorAlert($invoice);
+                }
             }
         }
 
@@ -141,7 +156,7 @@ class invoiceService {
         return $this->providerContactData[$invoiceNumber] ?? null;
     }
 
-    public function sendInvoiceToProvider($invoiceId)
+    public function sendInvoiceToProvider($invoiceId, $providerName=null)
     {
 
         $result = (object)[
@@ -159,15 +174,16 @@ class invoiceService {
 
                 $userEntity = new userEntity();
                 $userEntity->load($invoiceEntity->user_id);
+                $lookupContactBy = $providerName ? 'name' : 'email';
                 $contactData = (object)[
-                    'name' => trim($userEntity->first_name . ' ' . $userEntity->last_name),
+                    'name' => trim($providerName ?? $userEntity->first_name . ' ' . $userEntity->last_name),
                     'firstName' => $userEntity->first_name,
                     'lastName' => $userEntity->last_name,
                     'emailAddress' => $userEntity->email,
                     'taxNumber' => ""
                 ];
 
-                list($invoiceId, $invoiceNumber, $contactData) = $this->createProviderInvoice($lineItems, $contactData, 'email', null, '', $this->config->app->invoicing->xero->branding_themes->default ?? $this->config->app->invoicing->xero->default_branding_theme);
+                list($invoiceId, $invoiceNumber, $contactData) = $this->createProviderInvoice($lineItems, $contactData, $lookupContactBy, null, '', $this->config->app->invoicing->xero->branding_themes->default ?? $this->config->app->invoicing->xero->default_branding_theme);
 
                 $result->invoices_sent++;
 
@@ -184,7 +200,7 @@ class invoiceService {
                     $invoicePaymentEntity->loadByPaymentId($invoicePayment->id);
                     if (trim($invoicePaymentEntity->provider_payment_id == "")) {
                         $paymentEntity = new paymentEntity();
-                        $paymentEntity->load($invoicePaymentEntity->payment_id);
+                        $paymentEntity->load($invoicePaymentEntity->payment_id, true);
                         if ($paymentEntity->status == 'successful') {
                             $this->submitPaymentToProvider($paymentEntity, $invoiceEntity, $invoicePaymentEntity);
                             $result->invoice_payments_sent++;
@@ -229,12 +245,15 @@ class invoiceService {
         if ($quantity == 0) return [];
         $total = number_format($quantity*$perUnitCost, 2, '.', '');
         $totalRaw = round($quantity*$perUnitCost, 2);
+
+        $overrideNominalCode = System::getVariable('override_invoice_nominal_code') ?? null;
+
         //$nominalCode = '200';
         $lineItem = (object)[
             'description' => $description,
-            'accountCode' => $nominalCode,
+            'accountCode' => $overrideNominalCode ? $overrideNominalCode : $nominalCode,
             'quantity' => $quantity,
-            'perUnit' => !in_array($taxType, ['ECZROUTPUT', 'ZERORATEDOUTPUT']) ? $total : number_format(1.2*$totalRaw, 2, '.', ''),
+            'perUnit' => !in_array($taxType, ['ECZROUTPUT', 'ZERORATEDOUTPUT']) ? $perUnitCost : number_format(1.2*$perUnitCost, 2, '.', ''),
             'amount' => !in_array($taxType, ['ECZROUTPUT', 'ZERORATEDOUTPUT']) ? $total : number_format(1.2*$totalRaw, 2, '.', ''),
             'vat' => !in_array($taxType, ['NONE', 'ECZROUTPUT', 'ZERORATEDOUTPUT']) ? number_format(0.2*$totalRaw, 2, '.', '') : 0,
             'productId' => $productId
@@ -301,6 +320,8 @@ class invoiceService {
 
         if (!self::checkStartCron()) return System::asJson($response, ['error' => 'exit due to cron duplication. process: '.getmypid()]);
 
+        try {
+
         $subscriptionEntity = new subscriptionEntity();
         $productEntity = new productEntity();
         $subscriptionService = new subscriptionService();
@@ -372,6 +393,9 @@ class invoiceService {
             }
         }
         }
+        } catch (\Exception $e) {
+            $this->genericAlert('URGENT - GlobalM2MSIM subscription billing error', "There has been an error with the subscription invoicing process:<br><br> {$e->getMessage()}");
+        }
 
         return System::asJson($response, ['subscriptionsProcessed' => $subscriptionsProcessed]);
     }
@@ -385,6 +409,8 @@ class invoiceService {
     {
         if (!self::checkStartCron('invoice_payment')) return System::asJson($response, ['error' => 'exit due to cron duplication. process: '.getmypid()]);
 
+        try {
+
         $invoiceEntity = new invoiceEntity();
         $invoices = $invoiceEntity->getChargeable(1);
         $cardValidationService = new cardValidationService();
@@ -417,6 +443,9 @@ class invoiceService {
                 $invoiceEntity->last_payment_attempt = date('Y-m-d H:i:s'); // Set last payment attempt date and time
                 $invoiceEntity->store();
         }
+        } catch (\Exception $e) {
+            $this->genericAlert('URGENT - GlobalM2MSIM invoice payments error', "There has been an error with the invoice payment triggering process:<br><br> {$e->getMessage()}");
+        }
 
         return $result;
     }
@@ -476,6 +505,45 @@ class invoiceService {
         }
     }
 
+    protected function providerErrorAlert($invoiceEntity)
+    {
+        $userEntity = new userEntity();
+        $userEntity->load($invoiceEntity->user_id);
+
+        $emailData = [
+            'subject' => 'High Priority: GlobalM2MSIM Dashboard Invoice not sent to Provider',
+            'message' => "<p>Invoice Local ID: {$invoiceEntity->id} could not be sent to Xero.<br><br>Customer name: {$userEntity->getNiceName()}.</p><p>We need to investigate the reason for this, and ensure the invoice is billed correctly.</p>",
+        ];
+        $from = $this->config->app->site->email_address ?? 'no-reply@localhost.com';
+        $fromName = $this->config->app->site->name ?? 'localhost';
+        $mailer = new Mailer();
+        $mailer->setFrom($from, $fromName);
+        $mailer->addAddress("mike@developmentx.uk");     // Add a recipient
+
+        $mailer->Subject = 'High Priority: GlobalM2MSIM Dashboard Invoice not sent to Provider';
+        $mailer->Body = templateService::fetch("generic", $emailData);
+        $mailer->IsHTML(true);
+        $result = $mailer->send();
+    }
+
+    protected function genericAlert($subject, $message)
+    {
+        $emailData = [
+            'subject' => $subject,
+            'message' => "<p>$message</p>",
+        ];
+        $from = $this->config->app->site->email_address ?? 'no-reply@localhost.com';
+        $fromName = $this->config->app->site->name ?? 'localhost';
+        $mailer = new Mailer();
+        $mailer->setFrom($from, $fromName);
+        $mailer->addAddress("mike@developmentx.uk");     // Add a recipient
+
+        $mailer->Subject = $subject;
+        $mailer->Body = templateService::fetch("generic", $emailData);
+        $mailer->IsHTML(true);
+        $result = $mailer->send();
+    }
+
     protected static function checkStartCron($process='subscription')
     {
         $lastStarted = System::getVariable($process.'_cron_process_started');
@@ -581,7 +649,7 @@ class invoiceService {
 
     }
 
-    protected function getPaymentNotesFromInvoice($invoiceEntity)
+    public function getPaymentNotesFromInvoice($invoiceEntity)
     {
         $user = new userEntity();
         $user->load($invoiceEntity->user_id);
@@ -607,7 +675,7 @@ class invoiceService {
      * @param $userPresent
      * @return paymentEntity|null
      */
-    public function attemptInvoicePayment($invoiceEntity, $ref="", $isSubscriptionPayment=true, $data=[], $meta=null, $userPresent=false)
+    public function attemptInvoicePayment($invoiceEntity, $ref="", $isSubscriptionPayment=true, $data=[], $meta=null, $userPresent=false, $sendPaymentInfo=true)
     {
         $paymentEntity = null;
 
@@ -618,18 +686,20 @@ class invoiceService {
             $subscriptionEntity = new subscriptionEntity();
             $subscriptions = $subscriptionEntity->getByInvoiceId($invoiceEntity->id);
             $cardValidationService->setSubscriptions($subscriptions);
-            $billingToken = $paymentService->getBillingToken($invoiceEntity->user_id);
+            $billingToken = $paymentService->getBillingToken($invoiceEntity->user_id, $invoiceEntity->payment_provider);
                 $userEntity = new userEntity();
                 $userEntity->load($invoiceEntity->user_id);
 
             if (!$billingToken) { // No billable token found - place invoice on hold
                 $this->processSubscriptionInvoicePaymentFailure($invoiceEntity);
             } else {
-                $paymentEntity = $cardValidationService->capturePaymentWithToken($invoiceEntity->user_id, $invoiceEntity->payment_provider, ($invoiceEntity->total + $invoiceEntity->vat), $ref, $isSubscriptionPayment, $data, $meta, $userPresent);
+                $transactionData = [];
+                if ($invoiceEntity->transaction_data) $transactionData = json_decode($invoiceEntity->transaction_data);
+                $paymentEntity = $cardValidationService->capturePaymentWithToken($invoiceEntity->user_id, $invoiceEntity->payment_provider, ($invoiceEntity->total + $invoiceEntity->vat), $ref, $isSubscriptionPayment, $data, $meta, $transactionData, $userPresent);
                 if (isset($paymentEntity->id)) {
 
                     // Submit payment to Xero, or other provider
-                    if (trim($invoiceEntity->provider_id) != "") {
+                    if ($sendPaymentInfo && trim($invoiceEntity->provider_id) != "") {
                     $invoicePaymentEntity = $this->addPaymentToInvoice($invoiceEntity->id, $paymentEntity);
                     if (trim($invoiceEntity->provider_id) != "") {
                         if ($paymentEntity->status == 'successful') {
@@ -666,17 +736,25 @@ class invoiceService {
 
     public function reenableInvoiceSubscriptions($invoiceEntity, $reEnableSubscriptions = true)
     {
+        Logger::log('reenable_invoice_subscriptions_start', ['invoice_id' => $invoiceEntity->id, 'reEnableSubscriptions' => $reEnableSubscriptions]);
+
         $subscriptionEntity = new subscriptionEntity();
         $subscriptionService = new subscriptionService();
         $subscriptions = $subscriptionEntity->getByInvoiceId($invoiceEntity->id);
+
+        Logger::log('reenable_invoice_subscriptions_data', ['invoice_id' => $invoiceEntity->id, 'reEnableSubscriptions' => $reEnableSubscriptions, 'subscriptions' => $subscriptions]);
+
         if ($reEnableSubscriptions && is_array($subscriptions)) {
             foreach ($subscriptions as $subscription) {
-                if (trim($subscriptionEntity->cancellation_date) != "") {
+                Logger::log('reenable_invoice_subscriptions_in_loop', ['invoice_id' => $invoiceEntity->id, 'subscription' => $subscription]);
+                if ($subscription->enabled == '1' && trim($subscription->cancellation_date) != "") {
+                    Logger::log('reenable_invoice_subscriptions_in_if', ['invoice_id' => $invoiceEntity->id, 'subscription' => $subscription]);
                     $subscriptionEntity = new subscriptionEntity();
                     $subscriptionEntity->load($subscription->id);
-                    $subscriptionEntity->enabled = 1;
+                    //$subscriptionEntity->enabled = 1;
                     $subscriptionEntity->cancellation_date = null;
                     $subscriptionEntity->store();
+                    Logger::log('reenable_invoice_subscriptions_finished', ['invoice_id' => $invoiceEntity->id, 'subscriptionEntity' => $subscriptionEntity->getData()]);
                     Hook::processHook('subscription_enabled', $subscriptionEntity, "Subscription re-enabled due to successful payment for ".$invoiceEntity->invoice_number);
                 }
             }
@@ -1067,10 +1145,10 @@ class invoiceService {
      * @param $email
      * @return void
      */
-    public function getProviderInvoices($email)
+    public function getProviderInvoices($email, $expanded=false)
     {
         if ($this->provider) {
-            return $this->provider->getInvoices($email);
+            return $this->provider->getInvoices($email, $expanded);
         }
     }
 
@@ -1158,6 +1236,64 @@ class invoiceService {
             $this->provider->getContactHistory($identifer, $type);
         }
     }
+
+    public function changeInvoiceContact($invoiceId, $toContactName)
+    {
+        if ($this->provider) {
+            $this->provider->changeInvoiceContact($invoiceId, $toContactName);
+        }
+    }
+
+    public function backfillUserInvoices($userId)
+    {
+        $currentUser = currentUser::getCurrentUser();
+
+        $userEntity = new userEntity();
+        $userEntity->load($userId);
+        if ($userEntity->id ?? null) {
+
+            $existingInvoices = [];
+            $invoiceEntity = new invoiceEntity();
+            $invoices = $invoiceEntity->getForUser($userId);
+            foreach ($invoices as $inv) {
+                $existingInvoices[trim($inv->invoice_number)] = 1;
+            }
+
+            $remoteInvoices = $this->getProviderInvoices($userEntity->email, true);
+
+            $upserts = [];
+
+            foreach ($remoteInvoices as $remoteInvoice) {
+
+                if (!isset($existingInvoices[trim($remoteInvoice['name'])])) {
+
+                    $upserts[] = [
+                        'user_id' => $userId,
+                        'invoice_date' => $remoteInvoice['dateMysql'],
+                        'invoice_number' => trim($remoteInvoice['name']),
+                        'total' => str_replace(",","",$remoteInvoice['subtotal']),
+                        'vat' => str_replace(",","",$remoteInvoice['totalTax']),
+                        'line_items' => '',
+                        'payment_provider' => 'none',
+                        'provider_id' => $remoteInvoice['id'],
+                        'billing_enabled' => 0,
+                        'send_receipt' => 0,
+                        'created' => time(),
+                        'created_by' => $currentUser->id,
+                        'modified' => 0,
+                        'modified_by' => 0,
+                        'paid' => $remoteInvoice['paid'],
+                        'credit_applied' => '0.00',
+                        'provider' => $this->config->app->invoicing->provider ?? null
+                    ];
+                }
+            }
+
+            if (count($upserts) > 0) {
+                Database::persistRawBatch($upserts, 'invoice');
+            }
+        }
+    }
 }
 
 
